from functools import reduce

def build_config(transforms, config):
    """Reduces a dictionary through functions to modify the dictionary"""
    return reduce(lambda c, fn: fn(c), transforms, config)



import os, subprocess
import boto3
import click
from botocore.exceptions import ClientError
from .utils import err_log
from .service_config import ServiceConfig
from .deploy_file import DeployFile
from .task_config import TaskConfig
from airmail.utils.bash import run_script
from airmail.utils.files import read_package_json_version

class ECSService(DeployFile):
    def __init__(self, env=None, profile=None):
        # ECS client
        if profile is None:
            self.client = boto3.client('ecs')
        else:
            self.session = boto3.session.Session(profile_name=profile)
            self.client = self.session.client('ecs')
        # The env
        self.env = env
        self.profile = profile

        # Init the Deploy File class
        DeployFile.__init__(self, self.env)

        self.ServiceConfigBuilder = ServiceConfig(self.get_prop, self.get_top_level_prop, self.get_with_prefix)
        self.TaskConfigBuilder = TaskConfig(self.get_prop, self.get_top_level_prop, self.get_with_prefix)

    def get_env(self):
        """Return the env passed in"""
        return self.env

    # Check if a service exists
    def service_exists(self):
        """Given the config.yml file, check if the service exists in the target cluster"""

        cluster = self.get_top_level_prop('cluster')
        service_name = self.get_top_level_prop('service')

        try:
            # Make the request
            response = self.client.describe_services(
                cluster=cluster,
                services=[ service_name ]
            )

            # Grab the failures
            failures = response['failures']

            if len(failures) >= 1:
                return False
            else:
                return True

        except ClientError as e:
            if e.response['Error']['Code'] == 'ClusterNotFoundException':
                # If hitting this, the cluster does not exist
                err_log('Cluster not found!')
            else:
                click.echo('Unexpected error: ' + e)

    def create_service(self):
        """Create a service using the config.yml and default config (data/service.yml)"""
        # Build the config object
        config = self.ServiceConfigBuilder.build('service')
        response = self.client.create_service(**config)
        self.log_status(
            boto_response = response,
            success_msg='Service created!',
            err_msg='Service creation failed!'
        )

    def update_service(self):
        config = self.ServiceConfigBuilder.build('service_update')
        response = self.client.update_service(**config)
        self.log_status(
            boto_response = response,
            success_msg='Updated service!',
            err_msg='Service update failed!'
        )

    def create_task_definition(self):
        """Return the JSON for a task definition"""

        image_version = self.find_image_version()
        return self.TaskConfigBuilder.build(image_version, self.env)

    def build_container_image(self):
        """Build Container Image

        Executes a bash script to build an image with the
        Docker CLI and push that image to ECT
        """

        ecr_repo  = self.get_with_prefix('name', '/')
        version   = self.find_image_version()
        image_tag = self.get_top_level_prop('imageID') + ':' + version
        app_dir   = self.get_top_level_prop('projectDirectory')
        env_dict  = {
            'VERSION': version,
            'ECR_REPO': ecr_repo,
            'TASK_IMAGE_TAG': image_tag,
            'APP_DIR': app_dir
        }

        if not self.profile is None:
            env_dict['AWS_PROFILE'] = self.profile

        val = run_script('ecr_push', env_dict) # Returns the exit code from bash

        if (val == 0): # If a good run, return the tag
            return image_tag
        else: # TODO: Handle error
            print('The image build went poorly')

    def find_image_version(self):
        """Grabs the version of the image to build from the config file
        or pulls from the package.json file in the project directory
        """

        package_version = read_package_json_version(self.get_top_level_prop('projectDirectory'))
        config_version = self.get_prop('deployment.version')

        if config_version is not None:
            return config_version
        else:
            return package_version

    def register_task_definition(self, task):
        """Given task definition JSON, make the container and create a task definition"""

        response = self.client.register_task_definition(**task)
        self.log_status(
            boto_response = response,
            success_msg='Registered new task definition!',
            err_msg='Task definition registration failed!'
        )

    def log_status(self, boto_response, success_msg = "", err_msg = ""):
        if boto_response['ResponseMetadata']['HTTPStatusCode'] == 200:
            click.echo(click.style('Info: ', fg='green', bold=True) + success_msg)
        else:
            click.echo(click.style('Error: ', fg='red', bold=True) + err_msg)



from functools import reduce
from pydash import set_
from dotenv import dotenv_values
from airmail.utils.files import read_yml,determine_project_path
from airmail.utils.config import build_config

class TaskConfig():
    def __init__(self, get_prop, get_top_level_prop, get_with_prefix):
        # Getters passed in to read from the config file
        # TODO: cleanup later? Pass in a different away
        self.get_prop = get_prop
        self.get_top_level_prop = get_top_level_prop
        self.get_with_prefix = get_with_prefix

        # The list of functions to reduce through
        self.transforms = [
            self.assign_task_info,
            self.assignTaskRoles,
            self.assign_logging,
            self.assign_env_vars
        ]

    def build(self, version, env):
        # The config that will be sent to AWS client
        config = read_yml(determine_project_path() + '/../data/task.yml')
        # Assign the image version to use
        self.image_version = version
        # Assign the environment we're running in
        self.env = env
        # Run throught the config builder reduce
        return build_config(self.transforms, config)

    def assignTaskRoles(self, config):
        account_id = self.get_top_level_prop('accountID')
        config['taskRoleArn'] = 'arn:aws:iam::' + account_id + ':role/' + self.get_prop('taskRole')
        config['executionRoleArn'] = 'arn:aws:iam::' + account_id + ':role/' + self.get_prop('executionRole')
        return config

    def set_into_container_definition(self, config, prop, value):
        path = 'containerDefinitions[0].' + prop
        return set_(config, path, value)

    # Adds the desired count to the config
    def assign_task_info(self, config):
        config['family'] = self.get_top_level_prop('family')

        self.set_into_container_definition(config, 'name', self.get_top_level_prop('name'))
        self.set_into_container_definition(config, 'image', self.get_top_level_prop('imageID') + ':' + self.image_version)
        self.set_into_container_definition(config, 'command', self.get_prop('deployment.command'))
        self.set_into_container_definition(config, 'cpu', self.get_prop('deployment.cpu'))
        self.set_into_container_definition(config, 'memory', self.get_prop('deployment.memory'))
        self.set_into_container_definition(config, 'portMappings[0].containerPort', self.get_prop('deployment.port'))
        return config

    def assign_logging(self, config):
        """ Assign the logging configuration options to the task definition """
        self.set_into_container_definition(config, 'logConfiguration.logDriver', self.get_prop('logging.logDriver'))
        self.set_into_container_definition(config, 'logConfiguration.options', self.get_prop('logging.options'))
        return config

    def assign_env_vars(self, config):
        """ Read from the environment env file and turn the values into environment variables for the task definition """
        env_vars = []
        env_dict=dotenv_values(dotenv_path='./.deploy/' + self.env + '.env')

        for name, value in env_dict.items():
            env_vars.append({
                'name': name,
                'value': value
            })

        self.set_into_container_definition(config, 'environment', env_vars)
        return config




import os
import sys
import click

CONTEXT_SETTINGS = dict(auto_envvar_prefix='AIRMAIL')

class Context(object):
    def __init__(self):
        self.verbose = False
        self.home = os.getcwd()

    def log(self, msg, *args):
        """Logs a message to stderr."""
        if args:
            msg %= args
        click.echo(msg, file=sys.stderr)

    def vlog(self, msg, *args):
        """Logs a message to stderr only if verbose is enabled."""
        if self.verbose:
            self.log(msg, *args)

    def preamble_log(self, preamble, value):
        click.echo(click.style(preamble, fg='blue', bold=True) + ': ' + value)

    def info_log(self, msg):
        click.echo(click.style('Info', fg='green', bold=True) + ': ' + msg)

    def add_line_break(self):
        click.echo('\n')

    def err_log(self, msg, *args):
        if args:
            msg %= args
        click.echo(click.style('Error: ', fg='red', bold=True) + msg)


pass_context = click.make_pass_decorator(Context, ensure=True)
cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                          'commands'))


class AirMailCLI(click.MultiCommand):
    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith('.py') and \
               filename.startswith('cmd_'):
                rv.append(filename[4:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        try:
            if sys.version_info[0] == 2:
                name = name.encode('ascii', 'replace')
            mod = __import__('airmail.commands.cmd_' + name,
                             None, None, ['cli'])
        except ImportError:
            return
        return mod.cli

@click.command(cls=AirMailCLI, context_settings=CONTEXT_SETTINGS)
@click.option('-p', '--profile', help='The AWS profile to use', default=lambda: os.environ.get('AWS_PROFILE', None))
@click.option('-e', '--env', help='The environment to target', default=lambda: os.environ.get('ENV', ''))
@pass_context
def cli(ctx, profile, env):
    """A CLI for getting code into the cloud"""
    ctx.env = env
    ctx.profile = profile



import os, sys
from pathlib import Path
import click
import yaml

def err_log(msg, *args):
    if args:
        msg %= args
    click.echo(click.style('Error: ', fg='red', bold=True) + msg)

# Read a yaml file given a file path. Shou
def read_yml(filepath):
    target_file = Path(filepath)

    if target_file.is_file():
        with open(target_file) as yml_file:
            return yaml.load(yml_file)
    else:
        click.echo(click.style('Error: ', fg='red', bold=True) +
            'Could not read the .yml at ' +
            click.style(filepath, bold=True))
        sys.exit()

# Determines the path to files inside the
# project. (i.e. the `data` directory)
def determine_project_path():
    """Borrowed from wxglade.py"""
    try:
        root = __file__
        if os.path.islink (root):
            root = os.path.realpath (root)
        return os.path.dirname (os.path.abspath (root))
    except:
        print("I'm sorry, but something is wrong.")
        print("There is no __file__ variable. Please contact the author.")
        sys.exit ()



import os, sys
import click
from .utils import read_yml
from pydash import get, set_
import boto3

class DeployFile():
    def __init__(self, env):
        # Grab cwd
        self.cwd = os.getcwd()
        # Grab file
        self.file_path = self.cwd + '/.deploy/config.yml'
        # Read file or throw
        self.deploy_json = read_yml(self.file_path)
        # The environment
        self.env = env

        # Build some properties here
        # TODO: make it a reduce? :)
        self.inject_cluster_and_family()
        self.inject_aws_account_id()
        self.build_image_id()

    # Add top level account id
    def inject_aws_account_id(self):
        id = boto3.client('sts').get_caller_identity().get('Account')
        set_(self.deploy_json, 'accountID', id)

    # Grab the name from the file
    def get_name(self):
        return self.get_top_level_prop('name')

    # Grab the org name
    def get_org(self):
        return self.get_top_level_prop('org')

    def build_image_id(self):
        image_tag = self.get_top_level_prop('accountID') + '.dkr.ecr.us-east-1.amazonaws.com/' + self.get_with_prefix('name', '/')
        set_(self.deploy_json, '.imageID', image_tag)

    def get_with_prefix(self, prop, delim='-'):
        """Retrieve a value with <ORG>-<ENV>- prefix. Can pass in custom delimiter """
        top_level = self.get_top_level_prop(prop)
        value = top_level if top_level != None else self.get_prop(prop)
        return self.get_org() + delim + self.env + delim + value


    def inject_cluster_and_family(self):
        """Using the `name`, assign the cluster and family for the service/task"""

        cluster_val = self.get_with_prefix('name')
        set_(self.deploy_json, '.cluster', cluster_val)
        set_(self.deploy_json, '.family', cluster_val)

    #  Get a tol level property
    def get_top_level_prop(self, prop, default=None):
        """Retrieve a property's value from the top level of the config"""
        return get(self.deploy_json, prop, default)

    # Grab a property from the specific env
    def get_prop(self, prop, default=None):
        """Get a property's value that is nested in the env object"""
        prop = self.env + '.' + prop
        return get(self.deploy_json, prop, default)



import subprocess, os
from pathlib import Path
from .files import determine_project_path

def run_script(script, d):
    # Get the path from this file
    target_script = determine_project_path() + '/../bash/' + script + '.sh'
    # Execute the script
    my_env = os.environ.copy()

    for key, value in d.items():
        my_env[key] = value

    return subprocess.call([target_script], env=my_env)



from functools import reduce
from airmail.utils.files import read_yml,determine_project_path
from airmail.utils.config import build_config


class ServiceConfig():
    def __init__(self, get_prop, get_top_level_prop, get_with_prefix):
        # Getters passed in to read from the config file
        # TODO: cleanup later? Pass in a different away?
        self.get_prop = get_prop
        self.get_top_level_prop = get_top_level_prop
        self.get_with_prefix = get_with_prefix

        # The list of functions to reduce through
        self.transforms = [
            self.assign_load_balancer,
            self.assign_service_info,
            self.assign_deployment_configuration
        ]

    #
    def build(self, file):
        # The config that will be sent to AWS client
        config = read_yml(determine_project_path() + '/../data/' + file + '.yml')
        # Run throught the config builder reduce
        return build_config(self.transforms, config)

    # Adds the desired count to the config
    def assign_service_info(self, config):
        # update_service uses `service` and create_service uses `serviceName`
        service_prop = 'service' if 'service' in config else 'serviceName'

        config['cluster'] = self.get_top_level_prop('cluster')
        config['taskDefinition'] = self.get_top_level_prop('family')
        config['desiredCount'] = self.get_prop('deployment.desiredCount')
        config[service_prop] = self.get_top_level_prop('service')
        return config

    def assign_deployment_configuration(self, config):
        config['deploymentConfiguration']['maximumPercent'] = self.get_prop('deployment.maxHealthyPercent', 200)
        config['deploymentConfiguration']['minimumHealthyPercent'] = self.get_prop('deployment.minHealthyPercent', 80)
        return config

    # Adds load balancer information to the config
    def assign_load_balancer(self, config):
        # Add if there is a `loadBalancers` prop in the dictionary
        if 'loadBalancers' in config:
            config['loadBalancers'][0]['targetGroupArn'] = self.get_prop('loadBalancer.targetGroupArn')
            config['loadBalancers'][0]['containerName'] = self.get_top_level_prop('name')
            config['loadBalancers'][0]['containerPort'] = self.get_prop('deployment.port')
        return config

    # TODO: Add some retriever/setter/validation functions



import os
import click
import boto3
from airmail.cli import pass_context
from airmail.services.ecs_service import ECSService

@click.group(short_help='Create/modify ECS services')
@pass_context
def cli(ctx):
    if ctx.env is "":
        ctx.env = click.prompt('In which environment would you like to deploy?', type=str)

    ecs_service = ECSService(env=ctx.env, profile=ctx.profile)
    ctx.ecs_service = ecs_service
    ctx.add_line_break()
    ctx.preamble_log('Current Env', ecs_service.get_env())
    ctx.preamble_log('Project Name', ecs_service.get_name())
    ctx.preamble_log('Service Name', ecs_service.get_top_level_prop('service'))
    ctx.preamble_log('Project Cluster', ecs_service.get_top_level_prop('cluster'))
    ctx.preamble_log('Project Family', ecs_service.get_top_level_prop('family'))
    ctx.add_line_break()

@cli.command(name = 'deploy-service', short_help='Creates or updates a service based off the deploy configuration file')
@pass_context
def deploy_service(ctx):
    ctx.info_log('Running deploy-service...')
    ctx.info_log('This command will create an ECS service or update an existing service.')
    service_exists = ctx.ecs_service.service_exists() # Does the service exist?

    if service_exists:
        ctx.ecs_service.update_service()
    else:
        ctx.ecs_service.create_service()

@cli.command(name = 'register-task', short_help='Creates a task definition based off the deploy configuration file')
@pass_context
def register_task(ctx):
    ctx.info_log('Running register-task...')
    ctx.info_log('This command will create a task definition for project based off the deploy configuration file')
    ctx.add_line_break()
    task_definition = ctx.ecs_service.create_task_definition()
    ctx.ecs_service.register_task_definition(task_definition)


@cli.command(name = 'build-image', short_help='Builds a Docker image and pushes to ECR based on your deploy configuration')
@pass_context
def buildImage(ctx):
    ctx.info_log('Running build-image...')
    ctx.info_log('This command will build the image for ')
    ctx.add_line_break()

    built_tag = ctx.ecs_service.build_container_image()
    ctx.info_log('Built project image: ' + built_tag)
